//===========================================================================
//
//  Code written for CFA-10006 using an ATMEGA2561 running at 16 MHz.
//
//  CRYSTALFONTZ CFAL1602C-B CHARACTER OLED IN 8-BIT MODE
//
//  2015 - 10 - 14 Brent A. Crosby
//===========================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//===========================================================================
// This is supposed to be set somewhere in the AVR Studio, but I can't find
// the place :( In any case, our dev board, the CFA1006 runs at 16MHz, and
// this define makes the delays come out correctly.
# define F_CPU 16000000UL
//===========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

// control lines
#define  LCD_RESET PC2
#define  LCD_RD    PC5
#define  LCD_WR    PC6
#define  LCD_RS    PC7
#define  LCD_CS    PC4
#define LED_1     PD1
#define LED_2     PD2

// macros for control lines
#define CLR_RS PORTC &= ~(1<<LCD_RS);
#define SET_RS PORTC |=  (1<<LCD_RS);

#define CLR_CS PORTC &= ~(1<<LCD_CS);
#define SET_CS PORTC |=  (1<<LCD_CS);

#define CLR_RESET PORTC &= ~(1<<LCD_RESET);
#define SET_RESET PORTC |=  (1<<LCD_RESET);

#define CLR_RW PORTC &= ~(1 << LCD_WR);
#define SET_RW PORTC |=  (1 << LCD_WR);

#define CLR_E PORTC &= ~(1 << LCD_RD);
#define SET_E PORTC |=  (1 << LCD_RD);

#define LED1_OFF PORTD &= ~(1 << LED_1);
#define LED1_ON  PORTD |=  (1 << LED_1);

#define LED2_OFF PORTD &= ~(1 << LED_2);
#define LED2_ON  PORTD |=  (1 << LED_2);

#define DATA (1)
#define COMMAND (0)

//You can use port B.0 for timing with the oscilloscope
#define TIMIMG_MASK (0x01)
#define CLR_TIMIMG  (PORTB &= ~(TIMIMG_MASK))
#define SET_TIMING  (PORTB |=  (TIMIMG_MASK))
//===========================================================================
void wait_for_not_busy()
  {
  uint8_t
    busy_bit;
  uint16_t
    timeout;
  //Make Port A input
  DDRA=0x00;
  //Drive 0's out to turn off pull-ups
  PORTA=0x00;
  //Aim at the the control register
  CLR_RS;
  //Read it. Busy bit in position 7, address in bits 0-6
  SET_RW;
  //Enable the controller's data lines to drive the
  //port.
  SET_E;
  //Do not stay here forever if there is a hardware
  //problem.
  timeout=50000;
  do
    {
    busy_bit=PINA & (1<<PA7);
    }
  while((0x00 != busy_bit)&&(--timeout));
  //Get control of the bus back from the controller.
  CLR_E;
  //Default Port A back to output
  DDRA=0xFF;
  //If the busy line never went low, let Houston know
  //that we have a problem.
  if(0x00 == timeout)
    {
    //Use LED2 as a busy timeout error warning
    LED2_ON;
    }
  }
//===========================================================================
void blind_write_4_bit_command(uint8_t command)
  {
  //Put the high 4 bits of the data on the controller's data lines.
  PORTA=(command&0xF0) | (PORTA & 0x0F);
  //Aim at the the control register
  CLR_RS;
  //Writing the data
  CLR_RW;
  //Give a little bit of time for the control lines to settle
  _delay_us(1);
  //clock the data in
  SET_E;
  _delay_us(1);
  CLR_E;
  //Put the low 4 bits of the command data on the controller's data lines.
  PORTA=((command<<4)&0xF0) | (PORTA & 0x0F);
  //Give a little bit of time for the control lines to settle
  _delay_us(1);
  //clock the data in
  SET_E;
  _delay_us(1);
  CLR_E;
  }
//===========================================================================
void write_to_OLED_8_bit(uint8_t destination,uint8_t data)
  {
  //Assure that the controller has completed the previous command
  wait_for_not_busy();
  //Put the command data on the controller's data lines.
  PORTA=data;
  //Aim at the the control or data register
  if(DATA==destination)
    {
    SET_RS;
    }
  else
    {
    CLR_RS;
    }
  //Writing the data
  CLR_RW;
  //Give a little bit of time for the control lines to settle
  _delay_us(1);
  //clock the data in
  SET_E;
  _delay_us(1);
  CLR_E;
  }
//===========================================================================
void position_cursor(uint8_t column, uint8_t line)
  {
  //Set CGRAM Address, RS=0,R/W=0
  // 7654 3210
  // 1AAA AAAA
  //  0x00 to 0x27 => Line 1
  //  0x40 to 0x67 => Line 2
  write_to_OLED_8_bit(COMMAND,0x80 | (line?0x40:0x00) | (column & 0x3F));
  }
//===========================================================================
// If the controller has been initialized in graphic mode (how would that
// happen?) the regular character mode init is not good enough to bring it
// about.
//
// Of course it is very unlikely that the controller could randomly fall
// into graphics initialized mode, but it tells me that the regular
// character mode init is not as complete as I would like for a fully
// robust system.
//
// Adding this initialization code before the regular character mode init
// seems to bring it around nicely.
//---------------------------------------------------------------------------
void emulate_power_on_auto_initialization(void)
  {
  //In the (very) unlikely case that the display has already been initialized
  //in 4-bit mode, we need to do a 4-bit write to put it into 8-bit mode.
  //
  //Since I was testing this code across bot 4-bit and 8-bit versions, I
  //ran into this situation quite a bit. In pratice, it should be impossible
  //for your 8-bit lcd in an 8-bit system to accidentally get into 4-bit mode.
  _delay_us(500);
  blind_write_4_bit_command(0x30);
  _delay_us(500);

  //Power on sequence
  //1. Display clear
  //Display Clear RS=0,R/W=0
  // 7654 3210
  // 0000 0001
  write_to_OLED_8_bit(COMMAND,0x01);
  //2. Function set:
  //DL="1": 8-bit interface data
  //N="0": 1-line display
  //F="0": 5 x 8 dot character font
  //Function set, RS=0,R/W=0
  // 7654 3210
  // 001L NFFT
  //  L = Interface Length, 1=8-bit, 0 = 4-bit
  //  N = lines: N=1 is 2 lines
  //  F = Font: 0 = 5x8, 1 = 5x10
  //  FT = Font Table:
  //     FT=00 is English/Japanese ~"standard" for character LCDs
  //     FT=01 is Western European I fractions, circle-c some arrows
  //     FT=10 is English/Russian
  //     FT=11 is Western European II my favorite, arrows, Greek letters
  write_to_OLED_8_bit(COMMAND,0x30);

  //3. Power turn off
  //PWR=”0”
  //Graphic vs character mode setting, RS=0,R/W=0
  // 7654 3210
  // 0001 GP11
  //  G = Mode: 1=graphic, 0=character
  //  C = Power: 1=0n, 0=off
  write_to_OLED_8_bit(COMMAND,0x13);

  //4. Display on/off control:
  //D="0": Display off
  //C="0": Cursor off
  //B="0": Blinking off
  //Display On/Off Control RS=0,R/W=0
  // 7654 3210
  // 0000 1DCB
  //  D = Display On
  //  C = Cursor On
  //  B = Cursor Blink
  write_to_OLED_8_bit(COMMAND,0x08);

  //5. Entry mode set
  //I/D="0": Decrement by 1
  //S="0": No shift
  //Entry Mode Set RS=0,R/W=0
  // 7654 3210
  // 0000 01IS
  //  I = Increment/or decrement
  //  S = Shift(scroll) data on line
  write_to_OLED_8_bit(COMMAND,0x04);

  //6. Cursor/Display shift/Mode / Pwr
  //S/C=”0”, R/L=”1”: Shifts cursor position to the right
  //G/C=”0”: Character mode
  //Pwr=”1”: Internal DCDC power on
  // Cursor/Display shift/Mode / Pwr
  // 7654 3210
  // 0001 GP11
  //  G = Mode: 1=graphic, 0=character
  //  P = Power: 1=0n, 0=off
  write_to_OLED_8_bit(COMMAND,0x14);
  write_to_OLED_8_bit(COMMAND,0x17);
  }
//===========================================================================
void initialize_display()
  {
  //Default the control lines to something
  CLR_E;
  CLR_RS;
  SET_CS;
  PORTA=0x00;
  //Refer to WS0010 data sheet, page 21
  // https://www.crystalfontz.com/products/document/3176/WS0010.pdf

  //The display controller requests:
  //  "Wait for power stabilization 500ms:
  _delay_ms(500);

  //Function set, RS=0,R/W=0
  // 7654 3210
  // 001L NFFT
  //  L = Interface Length, 1=8-bit, 0 = 4-bit
  //  N = lines: N=1 is 2 lines
  //  F = Font: 0 = 5x8, 1 = 5x10
  //  FT = Font Table:
  //     FT=00 is English/Japanese ~"standard" for character LCDs
  //     FT=01 is Western European I fractions, circle-c some arrows
  //     FT=10 is English/Russian
  //     FT=11 is Western European II my favorite, arrows, Greek letters
  write_to_OLED_8_bit(COMMAND,0x3B);

  //Graphic vs character mode setting, RS=0,R/W=0
  // 7654 3210
  // 0001 GP11
  //  G = Mode: 1=graphic, 0=character
  //  C = Power: 1=0n, 0=off
  write_to_OLED_8_bit(COMMAND,0x17);

  //Display On/Off Control RS=0,R/W=0
  // 7654 3210
  // 0000 1DCB
  //  D = Display On
  //  C = Cursor On
  //  B = Cursor Blink
  write_to_OLED_8_bit(COMMAND,0x0E);

  //Display Clear RS=0,R/W=0
  // 7654 3210
  // 0000 0001
  write_to_OLED_8_bit(COMMAND,0x01);

  //Display home
  write_to_OLED_8_bit(COMMAND,0x02);

  //Entry Mode Set RS=0,R/W=0
  // 7654 3210
  // 0000 01IS
  //  I = Increment/or decrement
  //  S = Shift(scroll) data on line
  write_to_OLED_8_bit(COMMAND,0x06);

  // clear display
  write_to_OLED_8_bit(COMMAND,0x01);
  }
//===========================================================================
int main(void)
  {
  DDRA=0xFF;    // all pins as outputs
  PORTA=0xFF;   // data lines low
  DDRC=0xFF;    // all pins as outputs
  PORTC=0xFF;   // control and other lines high

  DDRD |= 0x06;  // Switches as inputs, LEDs as outputs
  PORTD=0xF0;   // pull ups on the keys, remaining pins low

  //Simple demo loop
  while (1)
    {
    //Hammer it back to the stone age
    emulate_power_on_auto_initialization();
    // Initialize the display
    initialize_display();

    //Display text screen
    //              0123456789012345
    char line1[17]=">}CRYSTALFONTZ{<";
    line1[ 0]=0xF6; // right triange;
    line1[ 1]=0xC7; // right triange;
    line1[14]=0xC8; // left triange;
    line1[15]=0xF7; // left triange;

    //              0123456789012345
    char line2[17]="OLED:aaaaaaaaaa8";
    line2[ 5]=0xE0; //beta
    line2[ 6]=0xA8; //function
    line2[ 7]=0xB8; //divide
    line2[ 8]=0xDA; //sigma
    line2[ 9]=0xEA; //mu (micro)
    line2[10]=0xDF; //alpha
    line2[11]=0xDE; //omega
    line2[12]=0xCF; //circle C
    line2[13]=0x1A; //about equal
    line2[14]=0xED; //pi

    // write to the first line
    position_cursor(0,0);
    for(int col=0;col<16;col++)
      {
      write_to_OLED_8_bit(DATA,line1[col]);
      }

    // write to the second line
    position_cursor(0,1);
    for(int col=0;col<16;col++)
      {
      write_to_OLED_8_bit(DATA,line2[col]);
      }

    //Show the message for a bit
    _delay_ms(1000);

    //Use the custom characters (CGRAM) to make the
    //left half solid, the right half checkerboard

    //Set CGRAM [0], to solid
    write_to_OLED_8_bit(COMMAND,0x40 | (0 <<3));
    for(int i=1;i<=8;i++)
    write_to_OLED_8_bit(DATA,0xFF);
    //Set CGRAM [1], checkerboard
    write_to_OLED_8_bit(COMMAND,0x40 | (1 <<3));
    for(int i=1;i<=4;i++)
      {
      write_to_OLED_8_bit(DATA,0xAA);
      write_to_OLED_8_bit(DATA,0x55);
      }

    // write to the first line
    position_cursor(0,0);
    for(int col=0;col<8;col++)
      {
      write_to_OLED_8_bit(DATA,0);  // CGRAM[0]
      }
    for(int col=8;col<16;col++)
      {
      write_to_OLED_8_bit(DATA,1);  // CGRAM[1]
      }

    // write to the second line
    position_cursor(0,1);
    for(int col=0;col<8;col++)
      {
      write_to_OLED_8_bit(DATA,0);  // CGRAM[0]
      }
    for(int col=8;col<16;col++)
      {
      write_to_OLED_8_bit(DATA,1);  // CGRAM[1]
      }

    //Show the pattern for a bit
    _delay_ms(1000);
    }  //while(1)
  }
//===========================================================================
