//===========================================================================
//
//  Code written for CFA-10006 using an ATMEGA2561 running at 16 MHz.
//
//  CRYSTALFONTZ CFAL1602C-B CHARACTER OLED IN SPI MODE
//
//  2015 - 10 - 14 Brent A. Crosby
//===========================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//===========================================================================
// This is supposed to be set somewhere in the AVR Studio, but I can't find
// the place :( In any case, our dev board, the CFA1006 runs at 16MHz, and
// this define makes the delays come out correctly.
#define F_CPU 16000000UL
//===========================================================================
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
//===========================================================================
// control lines
#define LED_1     PD1
#define LED_2     PD2

// macros for control lines
#define LED1_OFF PORTD &= ~(1 << LED_1);
#define LED1_ON  PORTD |=  (1 << LED_1);

#define LED2_OFF PORTD &= ~(1 << LED_2);
#define LED2_ON  PORTD |=  (1 << LED_2);
//===========================================================================
//SPI Mode, PORT B
// DB4 (0x10) is SS   (output) gray   pin 16
// DB5 (0x20) is SCK  (output) green  pin 12
// DB6 (0x40) is MISO (input)  violet pin 14
// DB7 (0x80) is MOSI (output) blue   pin 13
#define SPIPORT (PORTB)
#define SPITOGGLE (PINB)

//Blue = PB.4  LCD pin 13 MISO
//(reference only, it is an input)
#define MISO_MASK (0x10)
#define CLR_MISO (PORTB &= ~(MISO_MASK))
#define SET_MISO (PORTB |=  (MISO_MASK))

//Green = PB.5 LCD pin 12 SCK
#define SCK_MASK (0x20)
#define CLR_SCK (PORTB &= ~(SCK_MASK))
#define SET_SCK (PORTB |=  (SCK_MASK))

//Violet = PB.6 LCD pin 24  MOSI
#define MOSI_MASK (0x40)
#define CLR_MOSI (PORTB &= ~(MOSI_MASK))
#define SET_MOSI (PORTB |=  (MOSI_MASK))

//Gray = PB.7 LCD pin 16 SS
#define SS_MASK (0x80)
#define CLR_SS  (PORTB &= ~(SS_MASK))
#define SET_SS  (PORTB |=  (SS_MASK))

#define DATA (1)
#define COMMAND (0)

//You can use port B.0 for timing with the oscilloscope
#define TIMIMG_MASK (0x01)
#define CLR_TIMIMG  (PORTB &= ~(TIMIMG_MASK))
#define SET_TIMING  (PORTB |=  (TIMIMG_MASK))
//===========================================================================
void write_to_OLED_SPI(uint8_t destination,uint8_t command)
  {
  //Ideas taken from the fastest non-assembly implementation from
  //http://nerdralph.blogspot.com/2015/03/fastest-avr-software-spi-in-west.html
  //
  //SS low time ~ 3.8uS to 4.2uS for a 10-bit write
  //SCK freq ~ 2.4MHz
  //
  //Bits sent in this order:
  //  data(1)/command(0) flag
  //  read(1)/write(0) flag
  //  command.7 (msb)
  //  command.6
  //  command.5
  //  command.4
  //  command.3
  //  command.2
  //  command.1
  //  command.0 (lsb)

  //Pre-calculate the value that will drive clk and data low in one cycle.
  //(Note that this is not interrupt safe if an ISR is writing to the same port)
  register uint8_t
    force_clk_and_data_low;

  //Select the chip
  CLR_SS;

  //Pre-calculate the value that will drive clk and data low in one
  //operation.
  force_clk_and_data_low = (SPIPORT & ~(MOSI_MASK | SCK_MASK) );

  //Set clock and data low
  SPIPORT = force_clk_and_data_low;
  //Set the data(1)/command(0) flag
  if(DATA==destination)
    {
    SPITOGGLE = MOSI_MASK;
    }
  //Use a toggle to bring the clock up
  SPITOGGLE = SCK_MASK;

  //Set clock and data low
  SPIPORT = force_clk_and_data_low;
  //MOSI is already 0, for read(1)/write(0) flag as write
  //Use a toggle to bring the clock up
  SPITOGGLE = SCK_MASK;

  //Now clock the 8 bits of data out
  SPIPORT = force_clk_and_data_low;
  if(command & 0x80)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x40)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x20)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x10)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x08)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x04)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x02)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;
  SPIPORT = force_clk_and_data_low;
  if(command & 0x01)
    SPITOGGLE = MOSI_MASK;
  SPITOGGLE = SCK_MASK;

  //Release the chip
  SET_SS;

#if (0)
  //My original code, perhaps easier to understand.
  //SS low time is ~9.8uS
  //
  //Bits sent in this order:
  //  data(1)/command(0) flag
  //  read(1)/write(0) flag
  //  command.7 (msb)
  //  command.6
  //  command.5
  //  command.4
  //  command.3
  //  command.2
  //  command.1
  //  command.0 (lsb)
  //
  //Select the chip
  CLR_SS;

  //Set the data(1)/command(0) flag
  if(DATA==destination)
    SET_MOSI;
  else
    CLR_MOSI;
  //Clock it in.
  CLR_SCK; SET_SCK;

  //Clear the read(1)/write(0) flag to write, clock it in.
  CLR_MOSI;
  //Clock it in.
  CLR_SCK; SET_SCK;

  for(uint8_t mask=0x80;mask;mask>>=1)
    {
    mask&command?SET_MOSI:CLR_MOSI;
    CLR_SCK; SET_SCK;
    }

  //Release the chip
  SET_SS;
#endif
  }
//===========================================================================
void position_cursor(uint8_t column, uint8_t line)
  {
  //Set CGRAM Address, RS=0,R/W=0
  // 7654 3210
  // 1AAA AAAA
  //  0x00 to 0x27 => Line 1
  //  0x40 to 0x67 => Line 2
  write_to_OLED_SPI(COMMAND,0x80 | (line?0x40:0x00) | (column & 0x3F));
  }
//===========================================================================
// According to the WS0010 data sheet, only the clear display time has
// an appreciable execution time of 6mS. All others are listed at 0.
void clear_display(void)
  {
  //Display Clear RS=0,R/W=0
  // 7654 3210
  // 0000 0001
  write_to_OLED_SPI(COMMAND,0x01);
  _delay_ms(6);
  }
//===========================================================================
// If the controller has been initialized in graphic mode (how would that
// happen?) the regular character mode init is not good enough to bring it
// about.
//
// Of course it is very unlikely that the controller could randomly fall
// into graphics initialized mode, but it tells me that the regular
// character mode init is not as complete as I would like for a fully
// robust system.
//
// Adding this initialization code before the regular character mode init
// seems to bring it around nicely.
//---------------------------------------------------------------------------
void emulate_power_on_auto_initialization(void)
  {
  //Power on sequence
  //1. Display clear
  clear_display();

  //2. Function set:
  //DL="1": 8-bit interface data
  //N="0": 1-line display
  //F="0": 5 x 8 dot character font
  //Function set, RS=0,R/W=0
  // 7654 3210
  // 001D NFFT
  //  D = Data Length
  //  N = lines: N=1 is 2 lines
  //  F = Font: 0 = 5x8, 1 = 5x10
  //  FT = Font Table:
  //     FT=00 is English/Japanese ~"standard" for character LCDs
  //     FT=01 is Western European I fractions, circle-c some arrows
  //     FT=10 is English/Russian
  //     FT=11 is Western European II my favorite, arrows, Greek letters
  write_to_OLED_SPI(COMMAND,0x30);

  //3. Power turn off
  //PWR=0
  //Graphic vs character mode setting, RS=0,R/W=0
  // 7654 3210
  // 0001 GP11
  //  G = Mode: 1=graphic, 0=character
  //  C = Power: 1=0n, 0=off
  write_to_OLED_SPI(COMMAND,0x13);

  //4. Display on/off control:
  //D="0": Display off
  //C="0": Cursor off
  //B="0": Blinking off
  //Display On/Off Control RS=0,R/W=0
  // 7654 3210
  // 0000 1DCB
  //  D = Display On
  //  C = Cursor On
  //  B = Cursor Blink
  write_to_OLED_SPI(COMMAND,0x08);

  //5. Entry mode set
  //I/D="0": Decrement by 1
  //S="0": No shift
  //Entry Mode Set RS=0,R/W=0
  // 7654 3210
  // 0000 01IS
  //  I = Increment/or decrement
  //  S = Shift(scroll) data on line
  write_to_OLED_SPI(COMMAND,0x04);

  //6. Cursor/Display shift/Mode / Pwr
  //S/C=0, R/L=1: Shifts cursor position to the right
  //G/C=0: Character mode
  //Pwr=1: Internal DCDC power on
  // Cursor/Display shift/Mode / Pwr
  // 7654 3210
  // 0001 GP11
  //  G = Mode: 1=graphic, 0=character
  //  P = Power: 1=0n, 0=off
  write_to_OLED_SPI(COMMAND,0x14);
  write_to_OLED_SPI(COMMAND,0x17);
  }
//===========================================================================
void initialize_display()
  {
  //Refer to WS0010 data sheet, page 21
  // https://www.crystalfontz.com/products/document/3176/WS0010.pdf

  //The display controller requests:
  //  "Wait for power stabilization 500ms:
  _delay_ms(500);

  //Function set, RS=0,R/W=0
  // 7654 3210
  // 0011 NFFT
  //  N = lines: N=1 is 2 lines
  //  F = Font: 0 = 5x8, 1 = 5x10
  //  FT = Font Table:
  //     FT=00 is English/Japanese ~"standard" for character LCDs
  //     FT=01 is Western European I fractions, circle-c some arrows
  //     FT=10 is English/Russian
  //     FT=11 is Western European II my favorite, arrows, Greek letters
  write_to_OLED_SPI(COMMAND,0x3B);

  //Graphic vs character mode setting, RS=0,R/W=0
  // 7654 3210
  // 0001 GP11
  //  G = Mode: 1=graphic, 0=character
  //  C = Power: 1=0n, 0=off
  write_to_OLED_SPI(COMMAND,0x17);

  //Display On/Off Control RS=0,R/W=0
  // 7654 3210
  // 0000 1DCB
  //  D = Display On
  //  C = Cursor On
  //  B = Cursor Blink
  write_to_OLED_SPI(COMMAND,0x0E);

  //Display Clear RS=0,R/W=0
  // 7654 3210
  // 0000 0001
  clear_display();

  //Display home
  write_to_OLED_SPI(COMMAND,0x02);

  //Entry Mode Set RS=0,R/W=0
  // 7654 3210
  // 0000 01IS
  //  I = Increment/or decrement
  //  S = Shift(scroll) data on line
  write_to_OLED_SPI(COMMAND,0x06);

  //Display Clear RS=0,R/W=0
  // 7654 3210
  // 0000 0001
  clear_display();
  }
//===========================================================================
int main(void)
  {
  //General setup, port directions.
  DDRA=0xFF;    // all pins as outputs
  DDRB=0xFF;    // all pins as outputs
  DDRC=0xFF;    // all pins as outputs
  DDRD |= 0x06;  // Switches as inputs, LEDs as outputs

  PORTA=0xFF;   // data lines high
  PORTB=0xFF;   // data lines high
  PORTC=0xFF;   // control and other lines high
  PORTD=0xF0;   // pull ups on the keys, remaining pins low

  //Simple demo loop
  while (1)
    {
    //Hammer it back to the stone age
    emulate_power_on_auto_initialization();
    // Initialize the display
    initialize_display();

    //Display text screen
    //              0123456789012345
    char line1[17]=">}CRYSTALFONTZ{<";
    line1[ 0]=0xF6; // right triange;
    line1[ 1]=0xC7; // right triange;
    line1[14]=0xC8; // left triange;
    line1[15]=0xF7; // left triange;

    //              0123456789012345
    char line2[17]="OLED:aaaaaaaaaaS";
    line2[ 5]=0xE0; //beta
    line2[ 6]=0xA8; //function
    line2[ 7]=0xB8; //divide
    line2[ 8]=0xDA; //sigma
    line2[ 9]=0xEA; //mu (micro)
    line2[10]=0xDF; //alpha
    line2[11]=0xDE; //omega
    line2[12]=0xCF; //circle C
    line2[13]=0x1A; //about equal
    line2[14]=0xED; //pi

    //It takes 180uS to set the position twice and
    //write the 32 characters, so you could update
    //the screen memory at 5000 times per second :-O
    //In real life, the controller has a fixed
    //refresh rate of about ~140Hz.

    // write to the first line
    position_cursor(0,0);
    for(int col=0;col<16;col++)
      {
      write_to_OLED_SPI(DATA,line1[col]);
      }

    // write to the second line
    position_cursor(0,1);
    for(int col=0;col<16;col++)
      {
      write_to_OLED_SPI(DATA,line2[col]);
      }

    //Show the text message for a bit
    _delay_ms(1000);

    //Use the custom characters (CGRAM) to make the
    //left half solid, the right half checkerboard

    //Set CGRAM [0], to solid
    write_to_OLED_SPI(COMMAND,0x40 | (0 <<3));
    for(int i=1;i<=8;i++)
    write_to_OLED_SPI(DATA,0xFF);
    //Set CGRAM [1], checkerboard
    write_to_OLED_SPI(COMMAND,0x40 | (1 <<3));
    for(int i=1;i<=4;i++)
      {
      write_to_OLED_SPI(DATA,0xAA);
      write_to_OLED_SPI(DATA,0x55);
      }

    // write to the first line
    position_cursor(0,0);
    for(int col=0;col<8;col++)
      {
      write_to_OLED_SPI(DATA,0);  // CGRAM[0]
      }
    for(int col=8;col<16;col++)
      {
      write_to_OLED_SPI(DATA,1);  // CGRAM[1]
      }

    // write to the second line
    position_cursor(0,1);
    for(int col=0;col<8;col++)
      {
      write_to_OLED_SPI(DATA,0);  // CGRAM[0]
      }
    for(int col=8;col<16;col++)
      {
      write_to_OLED_SPI(DATA,1);  // CGRAM[1]
      }

    //Show the pattern for a bit
    _delay_ms(1000);
    }  //while(1)
  }
//===========================================================================
